@@ -1,16 +1,17 @@
+import itertools
+import logging
+import pathlib
+
+from pdfminer.layout import LAParams
+from pdfminer.pdfdocument import PDFDocument
+from pdfminer.pdfinterp import PDFResourceManager
+from pdfminer.pdfpage import PDFPage
+from pdfminer.pdfparser import PDFParser
+
 from .container import Container
 from .page import Page
 from .utils import resolve_and_decode
 
-import logging
-import pathlib
-import itertools
-from pdfminer.pdfparser import PDFParser
-from pdfminer.pdfdocument import PDFDocument
-from pdfminer.pdfpage import PDFPage
-from pdfminer.pdfinterp import PDFResourceManager
-from pdfminer.layout import LAParams
-
 logger = logging.getLogger(__name__)
 
 
@@ -22,14 +23,12 @@ class PDF(Container):
         stream,
         pages=None,
         laparams=None,
-        precision=0.001,
         password="",
         strict_metadata=False,
     ):
         self.laparams = None if laparams is None else LAParams(**laparams)
         self.stream = stream
         self.pages_to_parse = pages
-        self.precision = precision
         self.doc = PDFDocument(PDFParser(stream), password=password)
         self.rsrcmgr = PDFResourceManager()
         self.metadata = {}
