@@ -1,6 +1,7 @@
-from . import utils
-from operator import itemgetter
 import itertools
+from operator import itemgetter
+
+from . import utils
 
 DEFAULT_SNAP_TOLERANCE = 3
 DEFAULT_JOIN_TOLERANCE = 3
@@ -8,16 +9,20 @@ DEFAULT_MIN_WORDS_VERTICAL = 3
 DEFAULT_MIN_WORDS_HORIZONTAL = 1
 
 
-def snap_edges(edges, tolerance=DEFAULT_SNAP_TOLERANCE):
+def snap_edges(
+    edges, x_tolerance=DEFAULT_SNAP_TOLERANCE, y_tolerance=DEFAULT_SNAP_TOLERANCE
+):
     """
     Given a list of edges, snap any within `tolerance` pixels of one another
     to their positional average.
     """
-    v, h = [list(filter(lambda x: x["orientation"] == o, edges)) for o in ("v", "h")]
+    by_orientation = {"v": [], "h": []}
+    for e in edges:
+        by_orientation[e["orientation"]].append(e)
 
-    snap = utils.snap_objects
-    snapped = snap(v, "x0", tolerance) + snap(h, "top", tolerance)
-    return snapped
+    snapped_v = utils.snap_objects(by_orientation["v"], "x0", x_tolerance)
+    snapped_h = utils.snap_objects(by_orientation["h"], "top", y_tolerance)
+    return snapped_v + snapped_h
 
 
 def join_edge_group(edges, orientation, tolerance=DEFAULT_JOIN_TOLERANCE):
@@ -47,7 +52,9 @@ def join_edge_group(edges, orientation, tolerance=DEFAULT_JOIN_TOLERANCE):
     return joined
 
 
-def merge_edges(edges, snap_tolerance, join_tolerance):
+def merge_edges(
+    edges, snap_x_tolerance, snap_y_tolerance, join_x_tolerance, join_y_tolerance
+):
     """
     Using the `snap_edges` and `join_edge_group` methods above,
     merge a list of edges into a more "seamless" list.
@@ -59,16 +66,18 @@ def merge_edges(edges, snap_tolerance, join_tolerance):
         else:
             return ("v", edge["x0"])
 
-    if snap_tolerance > 0:
-        edges = snap_edges(edges, snap_tolerance)
+    if snap_x_tolerance > 0 or snap_y_tolerance > 0:
+        edges = snap_edges(edges, snap_x_tolerance, snap_y_tolerance)
 
-    if join_tolerance > 0:
-        _sorted = sorted(edges, key=get_group)
-        edge_groups = itertools.groupby(_sorted, key=get_group)
-        edge_gen = (
-            join_edge_group(items, k[0], join_tolerance) for k, items in edge_groups
+    _sorted = sorted(edges, key=get_group)
+    edge_groups = itertools.groupby(_sorted, key=get_group)
+    edge_gen = (
+        join_edge_group(
+            items, k[0], (join_x_tolerance if k[0] == "h" else join_y_tolerance)
         )
-        edges = list(itertools.chain(*edge_gen))
+        for k, items in edge_groups
+    )
+    edges = list(itertools.chain(*edge_gen))
     return edges
 
 
@@ -84,27 +93,31 @@ def words_to_edges_h(words, word_threshold=DEFAULT_MIN_WORDS_HORIZONTAL):
         return []
     min_x0 = min(map(itemgetter("x0"), rects))
     max_x1 = max(map(itemgetter("x1"), rects))
-    max_bottom = max(map(itemgetter("bottom"), rects))
-    edges = [
-        {
-            "x0": min_x0,
-            "x1": max_x1,
-            "top": r["top"],
-            "bottom": r["top"],
-            "width": max_x1 - min_x0,
-            "orientation": "h",
-        }
-        for r in rects
-    ] + [
-        {
-            "x0": min_x0,
-            "x1": max_x1,
-            "top": max_bottom,
-            "bottom": max_bottom,
-            "width": max_x1 - min_x0,
-            "orientation": "h",
-        }
-    ]
+
+    edges = []
+    for r in rects:
+        edges += [
+            # Top of text
+            {
+                "x0": min_x0,
+                "x1": max_x1,
+                "top": r["top"],
+                "bottom": r["top"],
+                "width": max_x1 - min_x0,
+                "orientation": "h",
+            },
+            # For each detected row, we also add the 'bottom' line.  This will
+            # generate extra edges, (some will be redundant with the next row
+            # 'top' line), but this catches the last row of every table.
+            {
+                "x0": min_x0,
+                "x1": max_x1,
+                "top": r["bottom"],
+                "bottom": r["bottom"],
+                "width": max_x1 - min_x0,
+                "orientation": "h",
+            },
+        ]
 
     return edges
 
@@ -408,7 +421,11 @@ DEFAULT_TABLE_SETTINGS = {
     "explicit_vertical_lines": [],
     "explicit_horizontal_lines": [],
     "snap_tolerance": DEFAULT_SNAP_TOLERANCE,
+    "snap_x_tolerance": None,
+    "snap_y_tolerance": None,
     "join_tolerance": DEFAULT_JOIN_TOLERANCE,
+    "join_x_tolerance": None,
+    "join_y_tolerance": None,
     "edge_min_length": 3,
     "min_words_vertical": DEFAULT_MIN_WORDS_VERTICAL,
     "min_words_horizontal": DEFAULT_MIN_WORDS_HORIZONTAL,
@@ -421,6 +438,24 @@ DEFAULT_TABLE_SETTINGS = {
     "intersection_y_tolerance": None,
 }
 
+NON_NEGATIVE_SETTINGS = [
+    "snap_tolerance",
+    "snap_x_tolerance",
+    "snap_y_tolerance",
+    "join_tolerance",
+    "join_x_tolerance",
+    "join_y_tolerance",
+    "edge_min_length",
+    "min_words_vertical",
+    "min_words_horizontal",
+    "text_tolerance",
+    "text_x_tolerance",
+    "text_y_tolerance",
+    "intersection_tolerance",
+    "intersection_x_tolerance",
+    "intersection_y_tolerance",
+]
+
 
 class TableFinder(object):
     """
@@ -434,20 +469,8 @@ class TableFinder(object):
     """
 
     def __init__(self, page, settings={}):
-        for k in settings.keys():
-            if k not in DEFAULT_TABLE_SETTINGS:
-                raise ValueError(f"Unrecognized table setting: '{k}'")
         self.page = page
-        self.settings = dict(DEFAULT_TABLE_SETTINGS)
-        self.settings.update(settings)
-        for var, fallback in [
-            ("text_x_tolerance", "text_tolerance"),
-            ("text_y_tolerance", "text_tolerance"),
-            ("intersection_x_tolerance", "intersection_tolerance"),
-            ("intersection_y_tolerance", "intersection_tolerance"),
-        ]:
-            if self.settings[var] is None:
-                self.settings.update({var: self.settings[fallback]})
+        self.settings = self.resolve_table_settings(settings)
         self.edges = self.get_edges()
         self.intersections = edges_to_intersections(
             self.edges,
@@ -457,6 +480,48 @@ class TableFinder(object):
         self.cells = intersections_to_cells(self.intersections)
         self.tables = [Table(self.page, t) for t in cells_to_tables(self.cells)]
 
+    @staticmethod
+    def resolve_table_settings(table_settings={}):
+        """Clean up user-provided table settings.
+
+        Validates that the table settings provided consists of acceptable keys and
+        returns a cleaned up version. The cleaned up version fills out the missing
+        values with the default values in the provided settings.
+
+        TODO: Can be further used to validate that the values are of the correct
+            type. For example, raising a value error when a non-boolean input is
+            provided for the key ``keep_blank_chars``.
+
+        :param table_settings: User-provided table settings.
+        :returns: A cleaned up version of the user-provided table settings.
+        :raises ValueError: When an unrecognised key is provided.
+        """
+        for k in table_settings.keys():
+            if k not in DEFAULT_TABLE_SETTINGS:
+                raise ValueError(f"Unrecognized table setting: '{k}'")
+
+        for setting in NON_NEGATIVE_SETTINGS:
+            if (table_settings.get(setting) or 0) < 0:
+                raise ValueError(f"Table setting '{setting}' cannot be negative")
+
+        resolved_table_settings = dict(DEFAULT_TABLE_SETTINGS)
+        resolved_table_settings.update(table_settings)
+
+        for var, fallback in [
+            ("text_x_tolerance", "text_tolerance"),
+            ("text_y_tolerance", "text_tolerance"),
+            ("snap_x_tolerance", "snap_tolerance"),
+            ("snap_y_tolerance", "snap_tolerance"),
+            ("join_x_tolerance", "join_tolerance"),
+            ("join_y_tolerance", "join_tolerance"),
+            ("intersection_x_tolerance", "intersection_tolerance"),
+            ("intersection_y_tolerance", "intersection_tolerance"),
+        ]:
+            if resolved_table_settings[var] is None:
+                resolved_table_settings.update({var: resolved_table_settings[fallback]})
+
+        return resolved_table_settings
+
     def get_edges(self):
         settings = self.settings
         for name in ["vertical", "horizontal"]:
@@ -492,8 +557,8 @@ class TableFinder(object):
             else:
                 v_explicit.append(
                     {
-                        "x0": utils.decimalize(desc),
-                        "x1": utils.decimalize(desc),
+                        "x0": desc,
+                        "x1": desc,
                         "top": self.page.bbox[1],
                         "bottom": self.page.bbox[3],
                         "height": self.page.bbox[3] - self.page.bbox[1],
@@ -526,8 +591,8 @@ class TableFinder(object):
                         "x0": self.page.bbox[0],
                         "x1": self.page.bbox[2],
                         "width": self.page.bbox[2] - self.page.bbox[0],
-                        "top": utils.decimalize(desc),
-                        "bottom": utils.decimalize(desc),
+                        "top": desc,
+                        "bottom": desc,
                         "orientation": "h",
                     }
                 )
@@ -547,10 +612,12 @@ class TableFinder(object):
 
         edges = list(v) + list(h)
 
-        if settings["snap_tolerance"] > 0 or settings["join_tolerance"] > 0:
-            edges = merge_edges(
-                edges,
-                snap_tolerance=settings["snap_tolerance"],
-                join_tolerance=settings["join_tolerance"],
-            )
+        edges = merge_edges(
+            edges,
+            snap_x_tolerance=settings["snap_x_tolerance"],
+            snap_y_tolerance=settings["snap_y_tolerance"],
+            join_x_tolerance=settings["join_x_tolerance"],
+            join_y_tolerance=settings["join_y_tolerance"],
+        )
+
         return utils.filter_edges(edges, min_length=settings["edge_min_length"])
