@@ -1,19 +1,18 @@
-from pdfminer.utils import PDFDocEncoding
-from pdfminer.psparser import PSLiteral
-from pdfminer.pdftypes import PDFObjRef
-from decimal import Decimal, ROUND_HALF_UP
-import numbers
-from operator import itemgetter
 import itertools
-from functools import lru_cache as cache
+from operator import itemgetter
+
+from pdfminer.pdftypes import PDFObjRef
+from pdfminer.psparser import PSLiteral
+from pdfminer.utils import PDFDocEncoding
 
 DEFAULT_X_TOLERANCE = 3
 DEFAULT_Y_TOLERANCE = 3
+DEFAULT_X_DENSITY = 7.25
+DEFAULT_Y_DENSITY = 13
 
 
 def cluster_list(xs, tolerance=0):
-    tolerance = decimalize(tolerance)
-    if tolerance == Decimal(0):
+    if tolerance == 0:
         return [[x] for x in sorted(xs)]
     if len(xs) < 2:
         return [[x] for x in sorted(xs)]
@@ -33,7 +32,6 @@ def cluster_list(xs, tolerance=0):
 
 
 def make_cluster_dict(values, tolerance):
-    tolerance = decimalize(tolerance)
     clusters = cluster_list(set(values), tolerance)
 
     nested_tuples = [
@@ -145,34 +143,6 @@ def resolve_all(x):
         return x
 
 
-@cache(maxsize=int(10e4))
-def _decimalize(v, q=None):
-    # Convert int-like
-    if isinstance(v, numbers.Integral):
-        return Decimal(int(v))
-
-    # Convert float-like
-    elif isinstance(v, numbers.Real):
-        if q is not None:
-            return Decimal(repr(v)).quantize(Decimal(repr(q)), rounding=ROUND_HALF_UP)
-        else:
-            return Decimal(repr(v))
-    else:
-        raise ValueError(f"Cannot convert {v} to Decimal.")
-
-
-def decimalize(v, q=None):
-    # If already a decimal, just return itself
-    if type(v) == Decimal:
-        return v
-
-    # If tuple/list passed, bulk-convert
-    if isinstance(v, (tuple, list)):
-        return type(v)(decimalize(x, q) for x in v)
-    else:
-        return _decimalize(v, q)
-
-
 def is_dataframe(collection):
     cls = collection.__class__
     name = ".".join([cls.__module__, cls.__name__])
@@ -193,31 +163,18 @@ def dedupe_chars(chars, tolerance=1):
     """
     key = itemgetter("fontname", "size", "upright", "text")
     pos_key = itemgetter("doctop", "x0")
-    t = decimalize(tolerance)
 
     def yield_unique_chars(chars):
         sorted_chars = sorted(chars, key=key)
         for grp, grp_chars in itertools.groupby(sorted_chars, key=key):
-            for y_cluster in cluster_objects(grp_chars, "doctop", t):
-                for x_cluster in cluster_objects(y_cluster, "x0", t):
+            for y_cluster in cluster_objects(grp_chars, "doctop", tolerance):
+                for x_cluster in cluster_objects(y_cluster, "x0", tolerance):
                     yield sorted(x_cluster, key=pos_key)[0]
 
     deduped = yield_unique_chars(chars)
     return sorted(deduped, key=chars.index)
 
 
-def collate_line(line_chars, tolerance=DEFAULT_X_TOLERANCE):
-    tolerance = decimalize(tolerance)
-    coll = ""
-    last_x1 = None
-    for char in sorted(line_chars, key=itemgetter("x0")):
-        if (last_x1 is not None) and (char["x0"] > (last_x1 + tolerance)):
-            coll += " "
-        last_x1 = char["x1"]
-        coll += char["text"]
-    return coll
-
-
 def objects_to_rect(objects):
     return {
         "x0": min(map(itemgetter("x0"), objects)),
@@ -243,6 +200,19 @@ def bbox_to_rect(bbox):
     return {"x0": bbox[0], "top": bbox[1], "x1": bbox[2], "bottom": bbox[3]}
 
 
+def merge_bboxes(bboxes):
+    """
+    Given a set of bounding boxes, return the smallest bounding box that
+    contains them all.
+    """
+    return (
+        min(map(itemgetter(0), bboxes)),
+        min(map(itemgetter(1), bboxes)),
+        max(map(itemgetter(2), bboxes)),
+        max(map(itemgetter(3), bboxes)),
+    )
+
+
 DEFAULT_WORD_EXTRACTION_SETTINGS = dict(
     x_tolerance=DEFAULT_X_TOLERANCE,
     y_tolerance=DEFAULT_Y_TOLERANCE,
@@ -260,13 +230,11 @@ class WordExtractor:
             if s not in DEFAULT_WORD_EXTRACTION_SETTINGS:
                 raise ValueError(f"{s} is not a valid WordExtractor parameter")
 
-            if s in ["x_tolerance", "y_tolerance"]:
-                val = decimalize(val)
-
             setattr(self, s, val)
 
     def merge_chars(self, ordered_chars):
         x0, top, x1, bottom = objects_to_bbox(ordered_chars)
+        doctop_adj = ordered_chars[0]["doctop"] - ordered_chars[0]["top"]
         upright = ordered_chars[0]["upright"]
 
         direction = 1 if (self.horizontal_ltr if upright else self.vertical_ttb) else -1
@@ -276,6 +244,7 @@ class WordExtractor:
             "x0": x0,
             "x1": x1,
             "top": top,
+            "doctop": top + doctop_adj,
             "bottom": bottom,
             "upright": upright,
             "direction": direction,
@@ -286,12 +255,12 @@ class WordExtractor:
 
         return word
 
-    def char_begins_new_word(self, current_chars, next_char):
+    def char_begins_new_word(self, current_chars, current_bbox, next_char):
         upright = current_chars[0]["upright"]
         intraline_tol = self.x_tolerance if upright else self.y_tolerance
         interline_tol = self.y_tolerance if upright else self.x_tolerance
 
-        word_x0, word_top, word_x1, word_bottom = objects_to_bbox(current_chars)
+        word_x0, word_top, word_x1, word_bottom = current_bbox
 
         return (
             (next_char["x0"] > word_x1 + intraline_tol)
@@ -302,19 +271,28 @@ class WordExtractor:
 
     def iter_chars_to_words(self, chars):
         current_word = []
+        current_bbox = None
 
         for char in chars:
             if not self.keep_blank_chars and char["text"].isspace():
                 if current_word:
                     yield current_word
                     current_word = []
+                    current_bbox = None
 
-            elif current_word and self.char_begins_new_word(current_word, char):
+            elif current_word and self.char_begins_new_word(
+                current_word, current_bbox, char
+            ):
                 yield current_word
                 current_word = [char]
+                current_bbox = obj_to_bbox(char)
 
             else:
                 current_word.append(char)
+                if current_bbox is None:
+                    current_bbox = obj_to_bbox(char)
+                else:
+                    current_bbox = merge_bboxes([current_bbox, obj_to_bbox(char)])
 
         if current_word:
             yield current_word
@@ -364,21 +342,108 @@ def extract_words(chars, **kwargs):
     return WordExtractor(**settings).extract(chars)
 
 
+def words_to_layout(
+    words,
+    x_density=DEFAULT_X_DENSITY,
+    y_density=DEFAULT_Y_DENSITY,
+    x_shift=0,
+    y_shift=0,
+    y_tolerance=DEFAULT_Y_TOLERANCE,
+    presorted=False,
+):
+    """
+    Given a set of word objects generated by `extract_words(...)`, return a
+    string that mimics the structural layout of the text on the page(s), using
+    the following approach:
+
+    - Sort the words by (doctop, x0) if not already sorted.
+
+    - Calculate the initial doctop for the starting page.
+
+    - Cluster the words by doctop (taking `y_tolerance` into account), and
+      iterate through them.
+
+    - For each cluster, calculate the distance between that doctop and the
+      initial doctop, in points, minus `y_shift`. Divide that distance by
+      `y_density` to calculate the minimum number of newlines that should come
+      before this cluster. Append that number of newlines *minus* the number of
+      newlines already appended, with a minimum of one.
+
+    - Then for each cluster, iterate through each word in it. Divide each
+      word's x0, minus `x_shift`, by `x_density` to calculate the minimum
+      number of characters that should come before this cluster.  Append that
+      number of spaces *minus* the number of characters and spaces already
+      appended, with a minimum of one. Then append the word's text.
+
+    Note: This approach currently works best for horizontal, left-to-right
+    text, but will display all words regardless of orientation. There is room
+    for improvement in better supporting right-to-left text, as well as
+    vertical text.
+    """
+    rendered = ""
+    words_sorted = words if presorted else sorted(words, itemgetter("doctop", "x0"))
+    doctop_start = words_sorted[0]["doctop"] - words_sorted[0]["top"]
+    for ws in cluster_objects(words_sorted, "doctop", y_tolerance):
+        y_dist = (ws[0]["doctop"] - (doctop_start + y_shift)) / y_density
+        newlines = rendered.count("\n")
+        rendered += "\n" * max(min(1, newlines), round(y_dist) - newlines)
+        line = ""
+        for word in sorted(ws, key=itemgetter("x0")):
+            x_dist = (word["x0"] - x_shift) / x_density
+            line += " " * max(min(1, len(line)), round(x_dist) - len(line))
+            line += word["text"]
+        rendered += line
+    return rendered
+
+
+def collate_line(line_chars, tolerance=DEFAULT_X_TOLERANCE, layout=False):
+    coll = ""
+    last_x1 = None
+    for char in sorted(line_chars, key=itemgetter("x0")):
+        if (last_x1 is not None) and (char["x0"] > (last_x1 + tolerance)):
+            coll += " "
+        last_x1 = char["x1"]
+        coll += char["text"]
+    return coll
+
+
 def extract_text(
     chars,
-    x_tolerance=DEFAULT_X_TOLERANCE,
-    y_tolerance=DEFAULT_Y_TOLERANCE,
+    layout=False,
+    x_density=DEFAULT_X_DENSITY,
+    y_density=DEFAULT_Y_DENSITY,
+    x_shift=0,
+    y_shift=0,
+    **kwargs,
 ):
     chars = to_list(chars)
     if len(chars) == 0:
-        return None
+        return ""
 
-    doctop_clusters = cluster_objects(chars, "doctop", y_tolerance)
+    if layout:
+        words = extract_words(chars, **kwargs)
+        y_tolerance = kwargs.get("y_tolerance", DEFAULT_Y_TOLERANCE)
+        return words_to_layout(
+            words,
+            x_density=x_density,
+            y_density=y_density,
+            x_shift=x_shift,
+            y_shift=y_shift,
+            y_tolerance=y_tolerance,
+            presorted=True,
+        )
 
-    lines = (collate_line(line_chars, x_tolerance) for line_chars in doctop_clusters)
+    else:
+        x_tolerance = kwargs.get("x_tolerance", DEFAULT_X_TOLERANCE)
+        y_tolerance = kwargs.get("y_tolerance", DEFAULT_Y_TOLERANCE)
 
-    coll = "\n".join(lines)
-    return coll
+        doctop_clusters = cluster_objects(chars, "doctop", y_tolerance)
+
+        lines = (
+            collate_line(line_chars, x_tolerance) for line_chars in doctop_clusters
+        )
+
+        return "\n".join(lines)
 
 
 collate_chars = extract_text
@@ -396,8 +461,8 @@ def filter_objects(objs, fn):
 
 
 def get_bbox_overlap(a, b):
-    a_left, a_top, a_right, a_bottom = decimalize(a)
-    b_left, b_top, b_right, b_bottom = decimalize(b)
+    a_left, a_top, a_right, a_bottom = a
+    b_left, b_top, b_right, b_bottom = b
     o_left = max(a_left, b_left)
     o_right = min(a_right, b_right)
     o_bottom = min(a_bottom, b_bottom)
@@ -418,7 +483,6 @@ def calculate_area(bbox):
 
 
 def clip_obj(obj, bbox):
-    bbox = decimalize(bbox)
 
     overlap = get_bbox_overlap(obj_to_bbox(obj), bbox)
     if overlap is None:
@@ -565,7 +629,7 @@ def rect_to_edges(rect):
     top.update(
         {
             "object_type": "rect_edge",
-            "height": decimalize(0),
+            "height": 0,
             "y0": rect["y1"],
             "bottom": rect["top"],
             "orientation": "h",
@@ -574,7 +638,7 @@ def rect_to_edges(rect):
     bottom.update(
         {
             "object_type": "rect_edge",
-            "height": decimalize(0),
+            "height": 0,
             "y1": rect["y0"],
             "top": rect["top"] + rect["height"],
             "doctop": rect["doctop"] + rect["height"],
@@ -584,7 +648,7 @@ def rect_to_edges(rect):
     left.update(
         {
             "object_type": "rect_edge",
-            "width": decimalize(0),
+            "width": 0,
             "x1": rect["x0"],
             "orientation": "v",
         }
@@ -592,7 +656,7 @@ def rect_to_edges(rect):
     right.update(
         {
             "object_type": "rect_edge",
-            "width": decimalize(0),
+            "width": 0,
             "x0": rect["x1"],
             "orientation": "v",
         }
